#include "cdi.def"
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>

//#include <iostream.h>
//#include <windows.h>
//#include <conio.h>
//#include "c:\\twincat\\ads api\\tcadsdll\\include\\tcadsdef.h"
//#include "c:\\twincat\\ads api\\tcadsdll\\include\\tcadsapi.h"
#include "tcadsdef.h"
#include "tcadsapi.h"

#include "tinecdi.h"
#include "cdi.h"
#include "twinCATadsLib.h"

#define N_TWINCAT_LINES 8
int nTwinCATLines = N_TWINCAT_LINES;
#define WAIT_TIME_NS 10000000  // 10ms
#define PLC_STRING_LENGTH	81
typedef struct
{
  int initFlag, cdiLine;
  long      nErr, nPort;  // local port number
  AmsAddr   Addr;         // remote SPS port & ip
  PAmsAddr  pAddr;
  char *tmpBuf;
  int	tmpLength;	// bytes
  int busInput[7];
} TWINCAT_DATA;
TWINCAT_DATA twincatInfo[N_TWINCAT_LINES] = { /* init */
  {0}, {0}, {0}, {0},  {0}, {0}, {0}, {0}
};

#define NUM_IGRPS 4		// can be 6, but last two access modes have not been tested yet!! 2010-May-3
int twinADSaccessGrp[] = {
    ADSIGRP_SYM_VALBYHND,
    ADSIGRP_MB,
    ADSIGRP_IOIMAGE_RWIB,	// read/write input byte(s)
    ADSIGRP_IOIMAGE_RWOB,	// read/write output byte(s)
	ADSIGRP_IOIMAGE_RWIX,	// read/write input bit
	ADSIGRP_IOIMAGE_RWOX,	// read/write output bit
};

char *scanCommand[] = {"getcommands","adsstatus","startplc","stopplc"};
enum scommand {getcommands, adsstatus, startplc, stopplc, nCommands};

int reqHandleTime=0, rwAdsTime=0, totalAdsTime=0;

__declspec(dllexport) int twinCatScanFcn(char *target,char *result,int resultBufferSize)
{
  char *c, lclTarget[256];
  int i, j, cnt, nPort, nErr, spsPort, slen = 0;
  int nData, data[10];
//  short rData, wData;
  short line = -1, crate = -1, subaddr = -1;
//  CdiLineInfoBlk *busHandle;
  AmsAddr   Addr;
  SHORT  nAdsState;
  USHORT  nDeviceState;

  if (result == NULL) return argument_list_error;
  if (resultBufferSize < 32) return buffer_too_small;
  // parse target according to "commandString:<line>:<crate>:<subaddr>"; e.g. NULL or "command:1" or "command:1:48" or "command:1:48:16" 
  if (target == NULL || strlen(target) == 0 ) target = "getcommands";
  strncpy(lclTarget,target,255);
  strlwr(lclTarget);
  // get command string
  if ((c=strtok(lclTarget,":")) == NULL) return 0;  // nothing to do
  for (j=0; j<nCommands; j++) if ( strcmp(scanCommand[j], c) == 0 ) break;
  if ( j == nCommands ) j = getcommands; // return the available commands as 'help'
  for ( nData=0; (c=strtok(NULL,":")) != NULL; nData++ ) data[nData] = strtol(c,NULL,0); //atoi(c);

  if (j == getcommands)
  {
    _snprintf(result,resultBufferSize,"availabled command strings:\n");
    for (i=0; i<nCommands; i++)
    {
      if ((slen = strlen(result)) > resultBufferSize - 12) break;
      strncat(result,scanCommand[i],resultBufferSize-slen);
      if (i <nCommands-1) strcat(result,",");
    }
    return 0;
  }

  if ( j == adsstatus || j == startplc || j == stopplc)
  {
    if ( nData > 7 || (nData < 4 && nData > 1) ) return argument_list_error;
    if ( j == adsstatus )
      _snprintf(result,resultBufferSize,"ADS Status:\n");
    else if ( j == startplc )
      _snprintf(result,resultBufferSize,"Start PLC:\n");
    else
      _snprintf(result,resultBufferSize,"Stop PLC:\n");
    nPort = AdsPortOpenEx();
    if ( nPort == 0 )
    {
      slen = strlen(result);
      strncat(result,"AdsPortOpenEx() failed!",resultBufferSize-slen);
      return 0;
    }
    if ( nData > 1 ) // open local address
    {
      cnt = nData < 6 ? 4 : 6;
      for (i=0; i<6; i++) Addr.netId.b[i] = (cnt == 6 || i < 4) ? data[i] : 1;
      spsPort = (nData % 2) ? data[cnt] : 0;
    }
    else
    {
      nErr = AdsGetLocalAddressEx(nPort, &Addr);
      if ( nErr )
      {
        if ((slen = strlen(result)) < resultBufferSize - 18) _snprintf(&result[slen],resultBufferSize-slen,"AdsGetLocalAddressEx failed: %d",nErr);
        AdsPortCloseEx(nPort);
        return 0;
      }
      spsPort = nData ? data[0] : 0;
    }
    cnt = spsPort <= 0 ? 4 : 1;   // check all 4 sps ports
    for ( i=0; i<cnt; i++)
    {
      if ( cnt == 1 ) Addr.port = spsPort;
      else
      {
        if ( i == 0 ) Addr.port = AMSPORT_R0_PLC_RTS1;
        else if ( i == 1 ) Addr.port = AMSPORT_R0_PLC_RTS2;
        else if ( i == 2 ) Addr.port = AMSPORT_R0_PLC_RTS3;
        else Addr.port = AMSPORT_R0_PLC_RTS4;
      }
      if ( j == adsstatus )
      {
        nErr = AdsSyncReadStateReqEx(nPort,(PAmsAddr)&Addr, &nAdsState, &nDeviceState);
        if ((slen = strlen(result)) < resultBufferSize - 18)
          if ( nErr )
            _snprintf(&result[slen],resultBufferSize-slen,"AdsSyncReadStateReq failed: %d; port: %d\n",nErr, Addr.port);
          else
            _snprintf(&result[slen],resultBufferSize-slen,"Port: %d; AdsState: %d; DeviceState: %d\n",Addr.port, nAdsState, nDeviceState);
      }
      else
      {
        nAdsState = j == startplc ? ADSSTATE_RUN : ADSSTATE_STOP;
        nDeviceState = 0;
        nErr = AdsSyncWriteControlReqEx(nPort,(PAmsAddr)&Addr, nAdsState, nDeviceState, 0, NULL);
        if ((slen = strlen(result)) < resultBufferSize - 18)
          if ( nErr )
            _snprintf(&result[slen],resultBufferSize-slen,"AdsSyncWriteControlReqEx failed: %d; port: %d\n",nErr, Addr.port);
          else
            _snprintf(&result[slen],resultBufferSize-slen,"Port: %d; AdsState: %d; DeviceState: %d\n",Addr.port, nAdsState, nDeviceState);
      }
    }

    nErr = AdsPortCloseEx(nPort);
    if ( nErr )
    {
      slen = strlen(result);
      strncat(result,"AdsPortCloseEx() failed!",resultBufferSize-slen);
      return 0;
    }
    return 0;
  }

  return 0;
}

__declspec(dllexport) int filterTwinCAT(int *parameter, int accessFlag)
{
  return 0;
}
#define CDI_PARAMETER_DELIMETER ":"
int cdiParseParamString(char *paramstr,int *parameters)
{
  char *s,*c=NULL;
  int i;
  if (paramstr == NULL || parameters == NULL) return argument_list_error;
  for (s=strtok_r(paramstr,CDI_PARAMETER_DELIMETER,&c), i=0; 
       i<NUMBER_EXEC_PARAM && s != NULL; 
       s=strtok_r(NULL,CDI_PARAMETER_DELIMETER,&c), i++)
  {
    parameters[i] = atoi(s);
  }
  return 0;
}
__declspec(dllexport) int initTwinCATcdi(int busLine, int cdiLine, int numberDevices, char *parameterList)
{
  cdi_mutex_t tmpMutex;
  int parameter[NUMBER_EXEC_PARAM];
  int lastError, error = 0;
 
  /* try to get Mutex, is this a first application  */
  lastError = cdiMutexInit(&tmpMutex, cdiLine);
//  if ( tmpMutex == NULL ) return -ask_error;
  if ( lastError == ERROR_ALREADY_EXISTS ) cdiMutexLock(&tmpMutex);

  if (busLine<1 || busLine>nTwinCATLines) error = -illegal_line; 
  memset(parameter,0,NUMBER_EXEC_PARAM*sizeof(int));
  cdiParseParamString(parameterList,parameter);
//  error = initTwinCAT(busLine);

  if ( lastError == ERROR_ALREADY_EXISTS ) cdiMutexUnlock(&tmpMutex);
  if ( error == 0 )
  {
	twincatInfo[busLine-1].cdiLine = cdiLine;
	twincatInfo[busLine-1].tmpBuf = NULL;
	twincatInfo[busLine-1].tmpLength = 0;
	memcpy(twincatInfo[busLine-1].busInput, &parameter[0], sizeof(int) * 7);
  }
  return error;
}

// if sync call, data not pass filter then repeat Transmit
__declspec(dllexport) int checkData(CdiRequestInfoBlk *pReq)
{
  struct timespec tv;

  /* check return condition */
  checkReturnCondition(pReq, (char *)pReq->bufLastData, pReq->pInData, pReq->inLength, 0);
  if ( pReq->dataReturnCondition == DATA_PASS_FILTER)
    memcpy( pReq->bufLastData, pReq->pInData, pReq->inLength < 8 ? pReq->inLength : 8);

 	if ( pReq->conditionFlag > anyData && pReq->dataReturnCondition != DATA_PASS_FILTER )
	{
    if ( pReq->callback == NULL )
    {
      if ( getTimeMilliseconds() - pReq->requestTime >= abs(pReq->timeout) )
      {
         pReq->errorCode = invalid_data;	// data back for sync call
         if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
      }
      else
      {
        tv.tv_sec = 0; tv.tv_nsec = WAIT_TIME_NS; // delay 10ms
        nanosleep( &tv, NULL );    
        return 1;
      }
    }
    pReq->active = finishRequest;
  }
  return 0;
}

int modifyAdsName(CdiRequestInfoBlk *pReq)
{
  char *pdest, *pdest2;
  char result[64];
  int ch, number, ret;

  // parse target astring: xxxx[]xxxxxx or xxxx[23]xxxxx
  ch ='[';
  pdest = strchr( pReq->pDev->lngName, ch );
  if ( pdest == NULL ) return 0;  // nothing to do

  ch =']';
  pdest2 = strchr( pReq->pDev->lngName, ch );
  if ( pdest2 == NULL ) return 0;  // nothing to do

  if ( pdest2 - pdest > 1 )
  {
    ret = sscanf(pdest,"[%d",&number);
    if ( ret == 1 ) return 0;   // nothing to do
  }
  strncpy(result,pReq->pDev->lngName, pdest - pReq->pDev->lngName);
  sprintf(&result[pdest - pReq->pDev->lngName],"[%d%s", pReq->pDev->cdiAddr[caAddrIndex], pdest2);
  strcpy(pReq->pDev->lngName, result);
  return 1;
}

int getAdsDeviceInfor(CdiRequestInfoBlk *pReq)
{ int ln, ret;
  ULONG lHdlVar, numBytesRead;
  AdsSymbolEntry AdsSymbolEntry;

  // user should set cdiAddr[subAddrIndex]: 0 or 1 (only these two values !!) 2007-8-17
  if ( pReq->pDev->cdiAddr[subAddrIndex] || pReq->pDev->devParameters[paraGroupIndex] ) return 0;
  ret = modifyAdsName(pReq);
  ln = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1;
  twincatInfo[ln].nErr = AdsSyncReadWriteReqEx2(twincatInfo[ln].nPort, twincatInfo[ln].pAddr, ADSIGRP_SYM_HNDBYNAME, 0x0, sizeof(lHdlVar), &lHdlVar, sizeof(pReq->pDev->lngName), pReq->pDev->lngName, &numBytesRead);
  printf("get symbole for: %s; value: %d; error number: %d\n",pReq->pDev->lngName, lHdlVar, twincatInfo[ln].nErr);
  if ( twincatInfo[ln].nErr ) cdilog("get symbole error for: %s; error: %d; line: %d\n",pReq->pDev->lngName, twincatInfo[ln].nErr, ln+1);
  else
  {
  	twincatInfo[ln].nErr = AdsSyncReadWriteReqEx2(twincatInfo[ln].nPort, twincatInfo[ln].pAddr,
								ADSIGRP_SYM_INFOBYNAMEEX,
								0x0,
								sizeof(AdsSymbolEntry),
								&AdsSymbolEntry, 
								sizeof(pReq->pDev->lngName),
								pReq->pDev->lngName, &numBytesRead);
	  if ( twincatInfo[ln].nErr )
  	{
	    printf("AdsSyncReadWriteReq(get group and index) error: %d; bus Line: %d\n",twincatInfo[ln].nErr, ln+1);
	    cdilog("AdsSyncReadWriteReq(get group and index) for: %s; error: %d; bus Line: %d\n",pReq->pDev->lngName, twincatInfo[ln].nErr, ln+1);
  	}
	  if ( twincatInfo[ln].nErr == 0 )
  	{
	    pReq->pDev->devParameters[paraGroupIndex] = lHdlVar; // store handle
	    pReq->pDev->devParameters[paraiGroup] = AdsSymbolEntry.iGroup;
	    pReq->pDev->devParameters[paraiOffs] = AdsSymbolEntry.iOffs;
	    pReq->pDev->devParameters[paraSize] = AdsSymbolEntry.size;
    }
  }
  return twincatInfo[ln].nErr;
}

int getTwincatIndexParameters(int *requestIndex, int numRequests, int reqSize, int ln, void **blockBuffer)
{ int i, size, reqTotalSize;
  TwinCatBlock *ptwinBlock;
  CdiRequestInfoBlk *pReq;

  reqTotalSize = reqSize + numRequests*4;
  size = sizeof(TwinCatBlock) + (sizeof(TDataPar) + sizeof(int)) * numRequests + reqTotalSize;
  if ( (ptwinBlock = calloc(size, sizeof(char))) == NULL )
  {
    cdilog("twinCatBlockTrans calloc mAdsSumBufferRes memory error: num Request: %d; size: %d\n", numRequests, reqTotalSize);
    return -out_of_local_memory;
  }
  ptwinBlock->reqList = (int *)((char *)ptwinBlock + sizeof(TwinCatBlock));
  ptwinBlock->twinCatMem = (TDataPar *)((char *)ptwinBlock->reqList + numRequests * sizeof(int));
  ptwinBlock->mAdsSumBufferRes = (unsigned char *)((char *)ptwinBlock->twinCatMem + numRequests * sizeof(TDataPar));
  ptwinBlock->numRequests = numRequests;
  ptwinBlock->reqSize = reqSize;
  ptwinBlock->reqTotalSize = reqTotalSize;
  *blockBuffer = (void *)ptwinBlock;
  for (i=0; i<ptwinBlock->numRequests; i++)   // store data
  {
    get_RequestBufferPointer(requestIndex[i], &pReq);
    ptwinBlock->reqList[i] = (int)pReq;
  	if ( pReq->pDev->cdiAddr[subAddrIndex] == 0 )
	  {
  	  ptwinBlock->twinCatMem[i].indexGroup  = pReq->pDev->devParameters[paraiGroup];
	    ptwinBlock->twinCatMem[i].indexOffset = pReq->pDev->devParameters[paraiOffs];
	    ptwinBlock->twinCatMem[i].length      = pReq->pDev->devParameters[paraSize];
  	}
	  else
  	{
	    ptwinBlock->twinCatMem[i].indexGroup  = twinADSaccessGrp[pReq->pDev->cdiAddr[subAddrIndex]];
	    if ( ptwinBlock->twinCatMem[i].indexGroup == ADSIGRP_MB ) ptwinBlock->twinCatMem[i].indexGroup += 0x20;	// should be 0x4040 ???
	    ptwinBlock->twinCatMem[i].indexOffset = pReq->pDev->cdiAddr[caAddrIndex] + pReq->pDev->devParameters[memOffset];
	    ptwinBlock->twinCatMem[i].length      = pReq->inLength;
  	}
//	ptwinBlock->reqSize += ptwinBlock->twinCatMem[i].length;
  }
  return 0;
}

__declspec(dllexport) int twinCatBlockTrans(int *requestIndex, int numberRequest, void **blockBuffer, int openFlag)
{ int i, retCode, busLine, cdiLine, access, ln, size;
  CdiRequestInfoBlk *pReq;

  if ( openFlag == 0  )  /* free buffer  */
  {
    if ( *blockBuffer == NULL || ((TwinCatBlock *)*blockBuffer)->reqList[0] == 0 )
    {
      cdilog("twinCatBlockTrans close Error! buffer pointer wrong !! buffer: 0x%x\n", *blockBuffer);
      return -invalid_parameter;
    }
    pReq = (CdiRequestInfoBlk *)((TwinCatBlock *)*blockBuffer)->reqList[0];
    if ( pReq->pBlockBuffer != *blockBuffer )
    {
      cdilog("twinCatBlockTrans close Error! buffer pointer wrong !! input buffer: 0x%x, allocated buffer: 0x%x, reqestIndex: %d\n", *blockBuffer, pReq->pBlockBuffer, pReq->requestNumber + 0x10000);
      return -invalid_parameter;
    }
    free(*blockBuffer);
    *blockBuffer = NULL;
    // cdilog("twinCatBlockTrans close OK! reqestIndex: %d\n", pReq->requestNumber + 0x10000);
    return 0;
  }
  *blockBuffer = NULL;
  size = 0;
  for (i=0; i<numberRequest; i++)   // check line
  {
    if ( (retCode = get_RequestBufferPointer(requestIndex[i], &pReq)) < 0 ) return retCode;
  	if ( pReq->pDev->cdiAddr[subAddrIndex] < 0 || pReq->pDev->cdiAddr[subAddrIndex] >= NUM_IGRPS ) return -invalid_parameter;
    if ( i == 0 )
    {
      access = pReq->accessFlag;
      if ( access != cdiReadFlag ) return -illegal_read_write;
      busLine = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine;
      cdiLine = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->cdiLine;
    }
	  if (busLine != ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine || cdiLine != ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->cdiLine ) return -illegal_line;
	  if ( access != pReq->accessFlag ) return -illegal_device_number;	/* device-access mode not consistent  */
	  ln = busLine - 1;
	  if ( twincatInfo[ln].initFlag == 0 )
  	{
	    twincatInfo[ln].nErr = -initTwinCAT(ln, pReq);
	    if ( twincatInfo[ln].nErr )
  	  {
	    	printf("initTwinCAT error: %d; bus Line: %d\n",twincatInfo[ln].nErr, ln+1);
		    return -hardware_error;
  	  }
	  }
	  // Fetch handle for an <szVar> PLC variable
	  twincatInfo[ln].nErr = getAdsDeviceInfor(pReq);
	  if ( twincatInfo[ln].nErr ) return -hardware_error;
	  // get total data size
	  size += pReq->pDev->cdiAddr[subAddrIndex] == 0 ? pReq->pDev->devParameters[paraSize] : pReq->inLength;
  }
//  ln = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1;
  retCode = getTwincatIndexParameters(requestIndex, numberRequest, size, ln, blockBuffer);
  if ( retCode ) return retCode > 0 ? (-hardware_error) : retCode;
  return 0;
}

void getSymbolAgain(CdiRequestInfoBlk *pReq, int i)
{ TwinCatBlock *ptwinBlock;

  if ( pReq->pDev->cdiAddr[subAddrIndex] ) return;
  pReq->pDev->devParameters[paraGroupIndex] = 0;
  if ( getAdsDeviceInfor(pReq) ) return;
  ptwinBlock = (TwinCatBlock *)pReq->pBlockBuffer;
  ptwinBlock->twinCatMem[i].indexGroup  = pReq->pDev->devParameters[paraiGroup];
  ptwinBlock->twinCatMem[i].indexOffset = pReq->pDev->devParameters[paraiOffs];
  ptwinBlock->twinCatMem[i].length      = pReq->pDev->devParameters[paraSize];
}

void blockTrans(CdiRequestInfoBlk *pReq)
{ int i, ln, numBytesRead, *pError;
  char *pData;
  TwinCatBlock *ptwinBlock;
  CdiRequestInfoBlk *tmpReq;

  if ( pReq->groupReq != pReq ) return;   // skip if not the first Group member
  ptwinBlock = (TwinCatBlock *)pReq->pBlockBuffer;
  for (i=0; i<ptwinBlock->numRequests; i++)   // reset rest of the request flag
  {
    tmpReq = (CdiRequestInfoBlk *)ptwinBlock->reqList[i];
    if ( tmpReq->groupReq != tmpReq)
    {
      tmpReq->errorCode = 0;
      if (tmpReq->perror != NULL) *tmpReq->perror = 0;
    }
  }
  ln = ((CdiLineInfoBlk *)tmpReq->pDev->pcdiLine)->busLine - 1;
  memset(ptwinBlock->mAdsSumBufferRes, 0, ptwinBlock->reqTotalSize);
  twincatInfo[ln].nErr = AdsSyncReadWriteReqEx2(twincatInfo[ln].nPort, twincatInfo[ln].pAddr, 
								0xf080, 
								ptwinBlock->numRequests, 
								ptwinBlock->reqTotalSize,	// we request additional "error"-flag(long) for each ADS-sub commands
								(void*)(ptwinBlock->mAdsSumBufferRes), 
								12*ptwinBlock->numRequests,	// send 12 bytes (3 * long : IG, IO, Len) of each ADS-sub command
								ptwinBlock->twinCatMem, &numBytesRead);
  pError = (int *)ptwinBlock->mAdsSumBufferRes;
  pData = (char *)&ptwinBlock->mAdsSumBufferRes[ptwinBlock->numRequests*4];
  for (i=0; i<ptwinBlock->numRequests; i++)   // set rest of the request data back flag
  {
    tmpReq = (CdiRequestInfoBlk *)ptwinBlock->reqList[i];
  	if ( twincatInfo[ln].nErr == 0 )	// copy data
	  {
  	  if ( tmpReq->perror ) *tmpReq->perror = pError[i] ? hardware_error : 0;
	    if ( pError[i] )
	    {
    		tmpReq->errorCode = hardware_error;
		    tmpReq->deviceError = twincatInfo[ln].nErr;
    		tmpReq->pDev->numberDeviceError++;
	    }
	    else memcpy(tmpReq->pInData, pData, tmpReq->inLength < ptwinBlock->twinCatMem[i].length ? tmpReq->inLength : ptwinBlock->twinCatMem[i].length);
  	}
	  else
  	{
	    tmpReq->errorCode = hardware_error;
	    tmpReq->deviceError = twincatInfo[ln].nErr;
	    tmpReq->pDev->numberDeviceError++;
	    if ( tmpReq->perror ) *tmpReq->perror = hardware_error;
      getSymbolAgain(tmpReq, i);
	  }
  	pData += ptwinBlock->twinCatMem[i].length;
    if ( twincatInfo[ln].nErr == 0 && pError[i] == 0 ) if ( tmpReq->hasDataMask ) checkData(tmpReq);
    if ( tmpReq->groupReq != tmpReq)
    {
      tmpReq->active = finishRequest; /* finish request */
      tmpReq->NgroupRets = 1;
    }
  }
}

__declspec(dllexport) void execTwinCAT(CdiRequestInfoBlk *pReq)
{
  if ( pReq->accessFlag > cdiWriteReadFlag )
  {
    pReq->errorCode = not_implemented;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    pReq->active = finishRequest; /* finish request */
    return;
  }

#ifdef TEST_TRUE
	*(char *)pReq->pInData = 0x2;
	printf("dev name: %s; hasMask: %d\n",pReq->pDev->devName, pReq->hasDataMask);
	printf("devParameters: 0x%x, 0x%x, 0x%x, 0x%x \n",
		pReq->pDev->cdiAddr[caAddrIndex], pReq->pDev->cdiAddr[subAddrIndex], pReq->pDev->devParameters[memOffset], pReq->pDev->devParameters[paraGroupIndex]);
    pReq->active = finishRequest; /* finish request */
    return;
#endif

  /* do first part; */
  cdiMutexLock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  if ( pReq->groupReq && pReq->pBlockBuffer ) blockTrans(pReq);
  else
  {

  do
  {
    doTwinCATRequest(pReq, 1);
    if ( pReq->active == finishRequest )
    {
      cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
      return;
    }
    /* check return condition */
    if (pReq->accessFlag != cdiReadFlag && pReq->accessFlag != cdiReadWriteFlag) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

  if ( pReq->active == finishRequest )
  {
    cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
    return;
  }
  if ( pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiWriteFlag)
  {
    pReq->active = finishRequest;
    cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
    return;
  }

  /* do second part */
  do
  {
    doTwinCATRequest(pReq, 2);
    if ( pReq->active == finishRequest )
    {
      cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
      return;
    }
    /* check return condition */
    if (pReq->accessFlag != cdiWriteReadFlag ) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

//  if (pReq->accessFlag == cdiWriteReadFlag )  printf("data: %d, %d, %d, %d, %d, %d\n",((int *)pReq->pInData)[0],((int *)pReq->pInData)[1],((int *)pReq->pInData)[2],((int *)pReq->pInData)[3],((int *)pReq->pInData)[4],((int *)pReq->pInData)[5]);
  }


  pReq->active = finishRequest; /* finish request */
  cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
}

// ln: 0 to  nTwinCATLines - 1
__declspec(dllexport) int initTwinCAT(int ln, CdiRequestInfoBlk *pReq)
{ int i, portNumber;

  // load twincat remote ip and port number
  if ( twincatInfo[ln].busInput[paraPortNumber] != 0 )
	  pReq->devParameters[paraPortNumber] = twincatInfo[ln].busInput[paraPortNumber];
  if ( twincatInfo[ln].busInput[paraNetId] != 0 )
	memcpy(&pReq->devParameters[paraNetId],&twincatInfo[ln].busInput[paraNetId],sizeof(int) * 6);

  printf("Twincat input devParameters: %d;%d; == %d; %d; %d, %d, %d, %d, %d\n",pReq->pDev->cdiAddr[caAddrIndex], pReq->pDev->cdiAddr[subAddrIndex], pReq->devParameters[0],pReq->devParameters[1],pReq->devParameters[2],pReq->devParameters[3],pReq->devParameters[4],pReq->devParameters[5],pReq->devParameters[6]);

  if ( pReq->pDev->cdiAddr[subAddrIndex] < 0 || pReq->pDev->cdiAddr[subAddrIndex] >= NUM_IGRPS ) return -invalid_parameter;

  // Open communication port for local PLC (Run-time system 1)
  if ( (twincatInfo[ln].nPort = AdsPortOpenEx()) == 0 )
  {
    printf("initTwinCAT error: AdsPortOpenEx() failed!; bus Line: %d\n", ln+1);
    return -not_ready;
  }
  // check netId input
  for (i=0; i<6; i++) if ( pReq->devParameters[paraNetId + i] ) break;
  if ( i != 6 ) // use input ip address
    for (i=0; i<6; i++) twincatInfo[ln].Addr.netId.b[i] = (char)pReq->devParameters[paraNetId + i];
  else
  {
    twincatInfo[ln].nErr = AdsGetLocalAddressEx(twincatInfo[ln].nPort, &(twincatInfo[ln].Addr));
    if ( twincatInfo[ln].nErr )
    {
      printf("initTwinCAT error: %d; bus Line: %d\n",twincatInfo[ln].nErr, ln+1);
      AdsPortCloseEx(twincatInfo[ln].nPort);
      return -not_ready;
    }
  }
  if ( pReq->devParameters[paraPortNumber] < 1 ) pReq->devParameters[paraPortNumber] = 1; // use first sps port
  if ( pReq->devParameters[paraPortNumber] >= 1 && pReq->devParameters[paraPortNumber] <= 4 ) // use PLC port number
  {
    if ( pReq->devParameters[paraPortNumber] == 1 ) portNumber = AMSPORT_R0_PLC_RTS1;
    else if ( pReq->devParameters[paraPortNumber] == 2 ) portNumber = AMSPORT_R0_PLC_RTS2;
    else if ( pReq->devParameters[paraPortNumber] == 3 ) portNumber = AMSPORT_R0_PLC_RTS3;
    else portNumber = AMSPORT_R0_PLC_RTS4;
  }
  else portNumber = pReq->devParameters[paraPortNumber];

  twincatInfo[ln].Addr.port = portNumber;

  twincatInfo[ln].pAddr = &(twincatInfo[ln].Addr);

  twincatInfo[ln].initFlag = 1;

  cdilog("Twincat portParameters: %d;%d; == %d; %d; %d, %d, %d, %d\n",twincatInfo[ln].nPort, portNumber, twincatInfo[ln].Addr.netId.b[0], twincatInfo[ln].Addr.netId.b[1],twincatInfo[ln].Addr.netId.b[2],twincatInfo[ln].Addr.netId.b[3],twincatInfo[ln].Addr.netId.b[4],twincatInfo[ln].Addr.netId.b[5]);
  printf("Twincat portParameters: %d;%d; == %d; %d; %d, %d, %d, %d\n",twincatInfo[ln].nPort, portNumber, twincatInfo[ln].Addr.netId.b[0], twincatInfo[ln].Addr.netId.b[1],twincatInfo[ln].Addr.netId.b[2],twincatInfo[ln].Addr.netId.b[3],twincatInfo[ln].Addr.netId.b[4],twincatInfo[ln].Addr.netId.b[5]);

  return 0;
}

__declspec(dllexport) int exitTwinCAT(int ln)
{
  CdiLineInfoBlk *busHandle;
  ULONG lHdlVar; 
  int i;

  if ( twincatInfo[ln].initFlag == 0 ) return 0;
  if ((busHandle=cdiGetBusHandle(twincatInfo[ln].cdiLine)) == NULL)
  {
  	cdilog("exitTwinCAT call cdiGetBusHandle error! cdiLine: %d, bus Line: %d\n",twincatInfo[ln].cdiLine, ln+1);
	  return -1;  /* something wrong!!  */
  }
  cdiMutexLock(&(busHandle->lineMutex));
  if ( twincatInfo[ln].tmpLength )
  {
  	free(twincatInfo[ln].tmpBuf);
	  twincatInfo[ln].tmpLength = 0;
  }
  for (i=0; i<busHandle->numberDevices; i++)
  	if ( busHandle->pDevList[i].cdiAddr[subAddrIndex] == 0 && busHandle->pDevList[i].pcdiLine && busHandle->pDevList[i].devParameters[paraGroupIndex] )
    {
  		lHdlVar = busHandle->pDevList[i].devParameters[paraGroupIndex];
//	    twincatInfo[ln].nErr = AdsSyncWriteReq(twincatInfo[ln].pAddr, ADSIGRP_SYM_RELEASEHND, 0, sizeof(lHdlVar), &lHdlVar); 
	    twincatInfo[ln].nErr = AdsSyncWriteReqEx(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_RELEASEHND, 0, sizeof(lHdlVar), &lHdlVar); 
	  	if ( twincatInfo[ln].nErr )  cdilog("exitTwinCAT: AdsSyncWriteReqEx(ADSIGRP_SYM_RELEASEHND) error: %d; bus Line: %d, deviceName: %s\n",twincatInfo[ln].nErr, ln+1, busHandle->pDevList[i].devName);
    }
  if ( twincatInfo[ln].nPort )
  {
	  twincatInfo[ln].nErr = AdsPortCloseEx(twincatInfo[ln].nPort);
	  twincatInfo[ln].nPort = 0;
  }
  if ( twincatInfo[ln].nErr )  cdilog("exitTwinCAT: AdsPortCloseEx error: %d; bus Line: %d\n",twincatInfo[ln].nErr, ln+1);
  else cdilog("ADSPortCloseEx OK!  cdiLine: %d, bus Line: %d\n",twincatInfo[ln].cdiLine, ln+1);

  twincatInfo[ln].initFlag = 0;
  cdiMutexUnlock(&(busHandle->lineMutex));
  return 0;
}

int waitTwinCATfinished()
{ int ln;
  for (ln=0; ln<N_TWINCAT_LINES; ln++) exitTwinCAT(ln);
  return 0;
}

__declspec(dllexport) int doTwinCATRequest(CdiRequestInfoBlk *pReq, int step)
{ int access, ln, locTime, secTime, size, i, j;
  ULONG lHdlVar, numBytesRead; 
  int tmpSize;

  if ( step == 1 ) /* first request */
    access = (pReq->accessFlag == cdiWriteFlag || pReq->accessFlag == cdiWriteReadFlag) ? CA_WRITE : CA_READ;
  else /* second */
    access = pReq->accessFlag == cdiReadWriteFlag ? CA_WRITE : CA_READ;

  //printf("try to get Mutex: process pointer: 0x%x; line: %d\n",pReq, pReq->pDev->cdiLine);
//  cdiMutexLock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  //printf("get Mutex: process pointer: 0x%x; line: %d\n",pReq, pReq->pDev->cdiLine);

  ln = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1;

  if ( twincatInfo[ln].initFlag == 0 )
  {
    twincatInfo[ln].nErr = -initTwinCAT(ln, pReq);
    if ( twincatInfo[ln].nErr )
    {
      printf("initTwinCAT error: %d; bus Line: %d\n",twincatInfo[ln].nErr, ln+1);
      pReq->errorCode = twincatInfo[ln].nErr;
      if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
      pReq->pDev->numberDeviceError++;
      pReq->active = finishRequest; /* finish request */
      return pReq->errorCode;
    }
  }

  // Fetch handle for an <szVar> PLC variable
  locTime = getTimeMilliseconds();
  twincatInfo[ln].nErr = getAdsDeviceInfor(pReq);
  if ( twincatInfo[ln].nErr == 0 )
	  if ( pReq->pDev->cdiAddr[subAddrIndex] == 0 ) lHdlVar = pReq->pDev->devParameters[paraGroupIndex];

  secTime = getTimeMilliseconds();
  if ( secTime - locTime > reqHandleTime )
  {
    reqHandleTime = secTime - locTime;
    cdilog("ADS reqHandleTime(ms): %d\n",reqHandleTime);
  }

  if ( twincatInfo[ln].nErr == 0 && pReq->pDev->dataFormat == cdiChar32 )
  {
  	size = (access == CA_READ) ? pReq->inLength : pReq->outLength;
	  size = size / getTypeLength(cdiChar32) * PLC_STRING_LENGTH;
  	if ( size > twincatInfo[ln].tmpLength )
	  {
  	  if ( twincatInfo[ln].tmpBuf ) free(twincatInfo[ln].tmpBuf);
	    if ( ( twincatInfo[ln].tmpBuf = calloc(size, sizeof(char)) ) == NULL )
  	  {
	    	twincatInfo[ln].tmpLength = 0;
		    twincatInfo[ln].nErr = out_of_local_memory;
	      printf("Twincat access mode: %d; callos == out of memory!!\n",access);
	      cdilog("Twincat access mode: %d; callos == out of memory!!\n",access);
  	  }
	    else twincatInfo[ln].tmpLength = size;
  	}
  }
  if ( twincatInfo[ln].nErr )
  {
    printf("AdsSyncReadWriteReq error: %d; bus Line: %d\n",twincatInfo[ln].nErr, ln+1);
    pReq->errorCode = hardware_error;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    pReq->pDev->numberDeviceError++;
    pReq->active = finishRequest; /* finish request */
  }
  else
  {
    if ( access == CA_READ )      // Read value of a PLC variable (by handle)
    {
      if ( pReq->pDev->cdiAddr[subAddrIndex] )
      {
	    	if ( pReq->pDev->dataFormat == cdiChar32 && size > PLC_STRING_LENGTH )
	        twincatInfo[ln].nErr = AdsSyncReadReqEx2(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, twinADSaccessGrp[pReq->pDev->cdiAddr[subAddrIndex]], pReq->pDev->cdiAddr[caAddrIndex] + pReq->pDev->devParameters[memOffset], size, (char *)twincatInfo[ln].tmpBuf, &numBytesRead );
    		else
		      twincatInfo[ln].nErr = AdsSyncReadReqEx2(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, twinADSaccessGrp[pReq->pDev->cdiAddr[subAddrIndex]], pReq->pDev->cdiAddr[caAddrIndex] + pReq->pDev->devParameters[memOffset], pReq->inLength, (char *)pReq->pInData, &numBytesRead );
//        twincatInfo[ln].nErr = AdsSyncReadReq(twincatInfo[ln].pAddr, twinADSaccessGrp[pReq->pDev->cdiAddr[subAddrIndex]], pReq->pDev->cdiAddr[caAddrIndex] + pReq->pDev->devParameters[memOffset], pReq->inLength, (char *)pReq->pInData);
//        twincatInfo[ln].nErr = AdsSyncReadReq(twincatInfo[ln].pAddr, pReq->pDev->cdiAddr[caAddrIndex], pReq->pDev->cdiAddr[subAddrIndex], pReq->inLength, (char *)pReq->pInData);
//        nErr = AdsSyncReadReq(pAddr, 0x4020, 0x0, 0x4, &dwData);
      }
      else if ( pReq->pDev->dataFormat >= 0 && (unsigned)getTypeLength((short)pReq->pDev->dataFormat) < pReq->inLength )
      {
    		if ( pReq->pDev->dataFormat == cdiChar32 && size > PLC_STRING_LENGTH )
	        twincatInfo[ln].nErr = AdsSyncReadReqEx2(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, size, (char *)twincatInfo[ln].tmpBuf, &numBytesRead );
		    else
		      twincatInfo[ln].nErr = AdsSyncReadReqEx2(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, pReq->inLength, (char *)pReq->pInData, &numBytesRead );
/*
        dataLen = getTypeLength((short)pReq->pDev->dataFormat);
        numRead = pReq->inLength / dataLen;
        for ( i=0; i<numRead; i++ )
          twincatInfo[ln].nErr = AdsSyncReadReqEx2(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, dataLen, ((char *)pReq->pInData) + dataLen * i, &numBytesRead );
*/
      }
	    else
  	  {
	    	if ( pReq->pDev->dataFormat == cdiChar32 && size > PLC_STRING_LENGTH )
	        twincatInfo[ln].nErr = AdsSyncReadReqEx2(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, size, twincatInfo[ln].tmpBuf, &numBytesRead );
    		else
		      twincatInfo[ln].nErr = AdsSyncReadReqEx2(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, pReq->inLength, pReq->pInData, &numBytesRead );
  	  }
#ifdef PRINT_RISTAU
		// print out for uwe Ristau
		if ( pReq->pDev->dataFormat == cdiChar32 && size > PLC_STRING_LENGTH ) tmpSize = size;
		else tmpSize = pReq->inLength;
		printf("returncode: %d; requested size: %d; readback size: %d\n",twincatInfo[ln].nErr,tmpSize,numBytesRead);
		cdilog("returncode: %d; requested size: %d; readback size: %d\n",twincatInfo[ln].nErr,tmpSize,numBytesRead);
		// print out for uwe Ristau
#endif
	    if ( twincatInfo[ln].nErr == 0 && pReq->pDev->dataFormat == cdiChar32 && size > PLC_STRING_LENGTH )
		    for(i=j=0; i<size; j++,i+=PLC_STRING_LENGTH)
			    memcpy(&((char *)pReq->pInData)[j*getTypeLength(cdiChar32)], (char *)&twincatInfo[ln].tmpBuf[i], getTypeLength(cdiChar32));
    }
    else
    {
      if ( pReq->pDev->dataFormat == cdiChar32 && size > PLC_STRING_LENGTH )
	     	for(i=j=0; i<size; j++,i+=PLC_STRING_LENGTH)
		      memcpy(&twincatInfo[ln].tmpBuf[i], &((char *)pReq->pOutData)[j*getTypeLength(cdiChar32)], getTypeLength(cdiChar32));
      if ( pReq->pDev->cdiAddr[subAddrIndex] )
      {
      	if ( pReq->pDev->dataFormat == cdiChar32 && size > PLC_STRING_LENGTH )
		      twincatInfo[ln].nErr = AdsSyncWriteReqEx(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, twinADSaccessGrp[pReq->pDev->cdiAddr[subAddrIndex]], pReq->pDev->cdiAddr[caAddrIndex] + pReq->pDev->devParameters[memOffset], size, (char *)twincatInfo[ln].tmpBuf);
    	 	else
		      twincatInfo[ln].nErr = AdsSyncWriteReqEx(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, twinADSaccessGrp[pReq->pDev->cdiAddr[subAddrIndex]], pReq->pDev->cdiAddr[caAddrIndex] + pReq->pDev->devParameters[memOffset], pReq->outLength, (char *)pReq->pOutData);
//        twincatInfo[ln].nErr = AdsSyncWriteReq(twincatInfo[ln].pAddr, twinADSaccessGrp[pReq->pDev->cdiAddr[subAddrIndex]], pReq->pDev->cdiAddr[caAddrIndex] + pReq->pDev->devParameters[memOffset], pReq->outLength, (char *)pReq->pOutData);
      }
      else if ( pReq->pDev->dataFormat >= 0 && (unsigned)getTypeLength((short)pReq->pDev->dataFormat) < pReq->outLength )
      {
		    if ( pReq->pDev->dataFormat == cdiChar32 && size > PLC_STRING_LENGTH )
    	    twincatInfo[ln].nErr = AdsSyncWriteReqEx(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, size, (char *)twincatInfo[ln].tmpBuf);
      	else
	        twincatInfo[ln].nErr = AdsSyncWriteReqEx(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, pReq->outLength, (char *)pReq->pOutData);
/*
        dataLen = getTypeLength((short)pReq->pDev->dataFormat);
        numRead = pReq->outLength / dataLen;
        for ( i=0; i<numRead; i++ )
          twincatInfo[ln].nErr = AdsSyncWriteReqEx(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, dataLen, ((char *)pReq->pOutData) + dataLen * i);
*/
      }
  	  else
  	  {
	  	  if ( pReq->pDev->dataFormat == cdiChar32 && size > PLC_STRING_LENGTH )
	        twincatInfo[ln].nErr = AdsSyncWriteReqEx(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, size, twincatInfo[ln].tmpBuf);
	    	else
  		    twincatInfo[ln].nErr = AdsSyncWriteReqEx(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_VALBYHND, lHdlVar, pReq->outLength, pReq->pOutData);
    	}
    }
    if ( twincatInfo[ln].nErr )
    {
      printf("Twincat access mode: %d; error: %d\n",access,twincatInfo[ln].nErr);
      pReq->errorCode = hardware_error;
      if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
      pReq->pDev->numberDeviceError++;
      pReq->active = finishRequest; /* finish request */
    }

    secTime = getTimeMilliseconds();
    if ( secTime - locTime > rwAdsTime )
    {
      rwAdsTime = secTime - locTime;
      cdilog("ADS reqHandleTime + rwTime (ms): %d\n",rwAdsTime);
    }

    //Handle der SPS-Variablen wieder freigeben
//2007-8-17
//    twincatInfo[ln].nErr = AdsSyncWriteReqEx(twincatInfo[ln].nPort,twincatInfo[ln].pAddr, ADSIGRP_SYM_RELEASEHND, 0, sizeof(lHdlVar), &lHdlVar); 
//	printf("exitTwinCAT: AdsSyncWriteReqEx(ADSIGRP_SYM_RELEASEHND) error: %d; bus Line: %d,  error number: %d\n",twincatInfo[ln].nErr, ln+1, twincatInfo[ln].nErr );

    secTime = getTimeMilliseconds();
    if ( secTime - locTime > totalAdsTime )
    {
      totalAdsTime = secTime - locTime;
      cdilog("ADS reqHandleTime + rwTime + closeHandle (ms): %d\n",totalAdsTime);
    }
  }
  // if twincat error, force next time load the handle again
  if ( twincatInfo[ln].nErr && pReq->pDev->cdiAddr[subAddrIndex] == 0 ) pReq->pDev->devParameters[paraGroupIndex] = 0;
//  cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  return pReq->errorCode;
}

