#ifndef CDIRS232_H_
#define CDIRS232_H_

#include "comm.def" 	// commport definition

#ifdef CDI_COMM_LINUX
#include <errno.h>
#include "termios.h"
#endif

CDI_COMM_EXPORT void execRS232(CdiRequestInfoBlk *pReq);
CDI_COMM_EXPORT int filterRS232Bus(int *parameter, int accessFlag);
CDI_COMM_EXPORT int initRS232ForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList);
CDI_COMM_EXPORT int closeSerialPort(int channelNumber);
CDI_COMM_EXPORT int commTimeoutFcn(CdiRequestInfoBlk *pReq, int set, int taskIndex);

#define N_SERIAL_LINES	8 	// com1 to com4 => change to com8 2011-Oct-5

#define STRING_LENGTH	255
#define dwInQueue     1200
#define dwOutQueue    1200

/* open serial parameter index */
enum openSerialParamIndex
{
  paramBaudrate, /* e.g 9600 */
  param8bits, /* flag 1 or 0... */
  param2StopBits,
  paramParity,
  paramOddParity,
  paramFlowControl,
  paramCanonicalInput,
  dataInterval = 10,
  bluetoothDevice,
};

enum ParamIndex
{
  index_1,
  index_2,
};

enum index6150AD
{
  dosisLeistung = 1,
  messSize,
  typInnenRohr,
  aktiveDetector
};

#define N_6150_DATA 4
typedef struct {
  int index, format, size;
} STRUCT_6150AD;
typedef struct {
  int index;
  char *detector;
} STRUCT_DETECTOR_6150AD;

typedef struct {
  int inputBaud, defBaud;
} BAUD_STRUCT;

#define REALLY_HAVE_RS232
#ifdef REALLY_HAVE_RS232
typedef struct {
#ifndef CDI_COMM_LINUX
  HANDLE serFd;		// serial File descripter
#else
  int serFd;		// serial File descripter
#endif
  int  channel,		// serial port channel, start from 0,1,2,3...
	cdiLine,		// cdi Line index for getString then callback
	baudRate,		// baudRate
	canonicalMode,
	usOneChar;	// time in us for one Char transmission
#ifndef CDI_COMM_LINUX
  HANDLE tIdMaster;	// receive task variables
  DCB dcbCopy;	// stored DCB
  COMMTIMEOUTS commtmoutsCopy;
  OVERLAPPED overlap[2];  // read and write
#else
  pthread_t	tIdMaster;	// receive task variables
  struct termios oldtio;	// stored tio
  fd_set fdset;
  struct timeval tout;
#endif
  char	buf[STRING_LENGTH + 1];	// input string
  
  int ret, length, wrtRet, recLength;
  unsigned int iOut;
  int timeStemp, interval, tmoutStemp, tmoutCounter;
  char	rcvData[STRING_LENGTH + 1];

  void (*receiveString)(int cdiLine, char *string, unsigned int length);
} SER_STRUCT;

extern SER_STRUCT serInfo[N_SERIAL_LINES];

extern BAUD_STRUCT baudRate[];
extern int numBaudRate;
extern int busDebugLevel;

#endif

#ifndef CDI_COMM_LINUX
#define COMM_LIB_NAME  "WINCOMM"
#else
#define COMM_LIB_NAME  "RS232"
#endif

#endif /* CDIRS232_H_ */
